/* logger.c: Logging Utilities
 *
 * Copyright (c) 2004-2010 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: Benjamin Li  (benli@broadcom.com)
 */

#include <errno.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <stdlib.h>

#include "options.h"
#include "logger.h"

/******************************************************************************
 * Default logger values
 ******************************************************************************/
static const char default_logger_filename[] = "/var/log/brcm-iscsi.log";

struct logger main_log = {
	.enabled	= LOGGER_ENABLED,
	.fp		= NULL,
	.log_file	= (char *) default_logger_filename,
	.level		= LOG_LEVEL_INFO,
	.lock		= PTHREAD_MUTEX_INITIALIZER,

	.stats		= {
		.debug	= 0,
		.info	= 0,
		.warn	= 0,
		.error	= 0,

		.last_log_time = 0,
	},
};

/******************************************************************************
 * Logger Functions
 ******************************************************************************/
/**
 *  log_uip() - Main logging function
 *  @param level_str - log level string
 *  @param fmt - log format
 */
void log_uip(char *level_str, char *fmt, ...)
{
	char time_buf[32];
	va_list ap, ap2;

	va_start(ap, fmt);

	if (main_log.fp == NULL) {
		goto end;
	}

	main_log.stats.last_log_time = time(NULL);
	strftime(time_buf, 26, "%a %b %d %T %Y",	
		 localtime(&main_log.stats.last_log_time));
	va_copy(ap2, ap);

	if (main_log.enabled == LOGGER_ENABLED) {
		fprintf(main_log.fp, "%s [%s]", level_str, time_buf);
		vfprintf(main_log.fp, fmt, ap);
		fprintf(main_log.fp, "\n");
	}

	if (opt.debug == DEBUG_ON) {
		fprintf(stdout, "%s [%s]", level_str, time_buf);
		vfprintf(stdout, fmt, ap2);
		fprintf(stdout, "\n");

		/* Force the printing of the log file */
		fflush(main_log.fp);

		/* Force the printing of the log out to standard output */
		fflush(stdout);
	}

end:
	va_end(ap2);
	va_end(ap);
}

/******************************************************************************
 *  Initialize/Clean up routines
 ******************************************************************************/
/**
 *  init_logger() - Prepare the logger
 *  @param filename - path to where the log will be written to
 *  @return 0 on success, <0 on failure
 */
int init_logger(char *filename)
{
	int rc = 0;

	pthread_mutex_lock(&main_log.lock);

	main_log.fp = fopen( filename, "a");
	if (main_log.fp == NULL) {
		printf("Could not create log file: %s <%s>\n",
		       filename, strerror(errno));
		rc = -EIO;
	}

	pthread_mutex_unlock(&main_log.lock);

	LOG_INFO("Initialize logger using log file: %s", filename);

	return rc;
}

void fini_logger(int type)
{
	pthread_mutex_lock(&main_log.lock);

	if (main_log.fp != NULL ) {
		fclose(main_log.fp);
		main_log.fp = NULL;

		if(opt.debug == DEBUG_ON) {
			printf("Closed logger\n");
			fflush(stdout);
		}
	}

	if (type == SHUTDOWN_LOGGER) {
		if ((main_log.log_file != NULL) && 
		    (main_log.log_file != default_logger_filename)) {	   
			free(main_log.log_file);
			main_log.log_file = NULL;
		}
	}

	main_log.enabled = LOGGER_DISABLED;

	pthread_mutex_unlock(&main_log.lock);
}
